/*
 * main.cxx
 *
 * PWLib application source file for Voxilla
 *
 * A H.323 "net telephone" application.
 *
 * Copyright (c) 1993-1998 Equivalence Pty. Ltd.
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.0 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific language governing rights and limitations
 * under the License.
 *
 * The Original Code is Portable Windows Library.
 *
 * The Initial Developer of the Original Code is Equivalence Pty. Ltd.
 *
 * Portions of this code were written with the assisance of funding from
 * Vovida Networks, Inc. http://www.vovida.com.
 *
 * Portions are Copyright (C) 1993 Free Software Foundation, Inc.
 * All Rights Reserved.
 *
 * Contributor(s): ______________________________________.
 *
 * $Log: main.cxx,v $
 * Revision 1.40  1999/08/31 12:34:19  robertj
 * Added gatekeeper support.
 *
 * Revision 1.39  1999/08/25 05:17:44  robertj
 * Simplified application structure (better for tutorial purposes)
 *
 * Revision 1.38  1999/08/10 00:56:33  robertj
 * Removed G723 option when not QuickNet
 *
 * Revision 1.37  1999/07/31 04:09:11  robertj
 * Added ability to create binary with no tracing code in it.
 *
 * Revision 1.36  1999/07/26 07:06:12  craigs
 * Added support for QuickNet Cards
 *
 * Revision 1.35  1999/07/19 01:59:53  robertj
 * Fixed crash on exit problem
 *
 * Revision 1.34  1999/07/18 11:00:59  robertj
 * Added user indication test.
 *
 * Revision 1.33  1999/07/16 16:05:14  robertj
 * Added more run time options.
 *
 * Revision 1.32  1999/07/16 14:05:10  robertj
 * Added application level jitter buffer adjustment.
 *
 * Revision 1.31  1999/07/16 02:12:56  robertj
 * Allow for disabling fast start initiation
 *
 * Revision 1.30  1999/07/14 06:05:37  robertj
 * GSM codec testing.
 *
 * Revision 1.29  1999/07/13 09:53:24  robertj
 * Fixed some problems with jitter buffer and added more debugging.
 *
 * Revision 1.28  1999/07/13 02:51:41  robertj
 * Fixed missing --sound parameter parsing.
 *
 * Revision 1.27  1999/07/10 03:45:32  robertj
 * Removed piece of ancient history (redundent code)
 *
 * Revision 1.26  1999/07/10 03:12:58  robertj
 * Moved version numbers into separate header file (pwlib build system compatibility)
 *
 * Revision 1.25  1999/07/10 02:59:26  robertj
 * Fixed ability to hang up incoming connection.
 *
 * Revision 1.24  1999/07/09 06:09:52  robertj
 * Major implementation. An ENORMOUS amount of stuff added everywhere.
 *
 * Revision 1.23  1999/06/25 17:51:56  robertj
 * New version
 *
 * Revision 1.22  1999/06/24 13:32:45  robertj
 * Fixed ability to change sound device on codec and fixed NM3 G.711 compatibility
 *
 * Revision 1.21  1999/06/22 13:42:31  robertj
 * Added user question on listener version to accept incoming calls.
 *
 * Revision 1.20  1999/06/14 15:35:03  robertj
 * New version number
 *
 * Revision 1.19  1999/06/14 08:00:29  robertj
 * Enhanced tracing again to add options to trace output (timestamps etc).
 *
 * Revision 1.18  1999/06/14 05:15:55  robertj
 * Changes for using RTP sessions correctly in H323 Logical Channel context
 *
 * Revision 1.17  1999/06/13 15:06:18  robertj
 * Added PConsoleChannel class for access to stdin/stdout/stderr.
 *
 * Revision 1.16  1999/06/13 12:41:13  robertj
 * Implement logical channel transmitter.
 * Fixed H245 connect on receiving call.
 *
 * Revision 1.15  1999/06/09 06:18:00  robertj
 * GCC compatibiltiy.
 *
 * Revision 1.14  1999/06/09 05:26:19  robertj
 * Major restructuring of classes.
 *
 * Revision 1.13  1999/06/07 01:00:11  robertj
 * Further debugging and error output
 *
 * Revision 1.12  1999/06/06 06:06:35  robertj
 * Changes for new ASN compiler and v2 protocol ASN files.
 *
 * Revision 1.11  1999/05/28 12:39:58  craigs
 * Added ability to select sound device at run time
 *
 * Revision 1.10  1999/05/22 15:22:34  craigs
 * Repaired damage to Windows sound interface
 *
 * Revision 1.9  1999/05/22 12:51:58  craigs
 * Fixed to work with new Linux implementation
 *
 * Revision 1.8  1999/05/22 05:00:45  craigs
 * Fixed stupid mistake with Sleep call
 *
 * Revision 1.7  1999/04/26 06:14:46  craigs
 * Initial implementation for RTP decoding and lots of stuff
 * As a whole, these changes are called "First Noise"
 *
 * Revision 1.6  1999/02/25 04:20:12  robertj
 * Added ability to redirect trace output
 *
 * Revision 1.5  1999/02/23 11:04:28  robertj
 * Added capability to make outgoing call.
 *
 * Revision 1.4  1999/01/16 01:31:35  robertj
 * Major implementation.
 *
 * Revision 1.3  1999/01/02 04:01:01  robertj
 * Added higher level protocol negotiations.
 *
 * Revision 1.2  1998/12/14 10:15:22  robertj
 * Fixed leak of sockets
 *
 * Revision 1.1  1998/12/14 09:13:19  robertj
 * Initial revision
 *
 */

#define	QUKCKNET

#include <ptlib.h>

#include "main.h"

#ifdef QUICKNET
#include "xjack.h"
#endif

#include "version.h"



PCREATE_PROCESS(Voxilla);

#define new PNEW


///////////////////////////////////////////////////////////////

Voxilla::Voxilla()
  : PProcess("Open H323 Project", "Voxilla",
             MAJOR_VERSION, MINOR_VERSION, BUILD_TYPE, BUILD_NUMBER)
{
}


Voxilla::~Voxilla()
{
}


void Voxilla::Main()
{
  cout << GetName()
       << " Version " << GetVersion(TRUE)
       << " by " << GetManufacturer()
       << " on " << GetOSClass() << ' ' << GetOSName()
       << " (" << GetOSVersion() << '-' << GetOSHardware() << ")\n\n";

  PArgList & args = GetArguments();
  args.Parse("l-listen."
#ifdef PTRACING
             "t-trace."
             "o-output:"
#endif
             "f-fast-disable."
             "g-gatekeeper:"
             "-no-gatekeeper."
             "-require-gatekeeper."
             "p-proxy:"
             "u-user:"
             "j-jitter:"
             "s-sound:"
             "-sound-in:"
             "-sound-out:"
             "-gsm."
             "-g711-ulaw."
             "-g711-Alaw."
#ifdef QUICKNET
             "-g723-63k."
             "q-quicknet:"
#endif
          , FALSE);

#ifdef PTRACING
  PTrace::SetOptions(PTrace::Timestamp | PTrace::Thread);
  PTrace::SetLevel(args.GetOptionCount('t'));
  if (args.HasOption('o')) {
    /* This is a bit tricky. We allocate the file for tracing on the heap, and
       we never actually deallocate it. So it come up as a leak on exit if
       memory checking is on. We cannot deallocate it as during termination
       some PTRACE's may get executed and attempt to trace to a closed (and
       deleted!) file object which crashes. Not good.
     */
    PTextFile * output = new PTextFile;
    if (output->Open(args.GetOptionString('o'), PFile::WriteOnly))
      PTrace::SetStream(output);
    else {
      cout << "Warning: could not open trace output file \""
           << args.GetOptionString('o') << '"' << endl;
      delete output;
    }
  }
#endif

  if (!args.HasOption('l') && args.GetCount() == 0) {
    cout << "Usage : " << GetName() << " [options] -l\n"
            "      : " << GetName() << " [options] [-p host] hostname\n"
            "Options:\n"
            "  -t --trace          : Enable trace, use multiple times for more detail\n"
            "  -o --output         : File for trace output, default is stderr\n"
            "  -f --fast-disable   : Disable fast start\n"
            "  -g --gatekeeper host: Specify gatekeeper host.\n"
            "  --no-gatekeeper     : Disable gatekeeper discovery.\n"
            "  --require-gatekeeper: Exit if gatekeeper discovery fails.\n"
            "  -f --fast-disable   : Disable fast start\n"
            "  -p --proxy host     : Proxy/Gateway hostname/ip address\n"
            "  -u --user name      : Set local user name (defaults to login name)\n"
            "  -j --jitter delay   : Set jitter buffer to delay milliseconds\n"
            "  -s --sound device   : Select sound input/output device\n"
            "  --sound-in device   : Select sound input device\n"
            "  --sound-out device  : Select sound output device\n"
            "  --gsm               : Set GSM 06.10 as preferred codec (default)\n"
            "  --g711-ulaw         : Set G.711 uLaw as preferred codec\n"
            "  --g711-Alaw         : Set G.711 ALaw as preferred codec\n"
#ifdef QUICKNET
            "  --g723-63k          : Set G.723.1 at 6.3k as preferred codec\n"
            "  -q -quicknet n      : use QuickNet device /dev/ixjn\n"
#endif
            "  -l --listen         : Listen for incoming calls\n"
            "  hostname/alias      : Remote host/alias to call\n";
    return;
  }

  MyH323EndPoint * endpoint;
#ifdef QUICKNET
  if (args.HasOption('q'))
    endpoint = new XJackH323EndPoint;
  else
#endif
    endpoint = new MyH323EndPoint;

  if (endpoint->Initialise(args)) {
    if (args.HasOption('l'))
      endpoint->ListenForIncomingCalls();
    else
      endpoint->MakeOutgoingCall(args[0], args.GetOptionString('p'));
    endpoint->AwaitTermination();
  }

  delete endpoint;
}


///////////////////////////////////////////////////////////////

BOOL MyH323EndPoint::Initialise(PArgList & args)
{
  if (!SetSoundDevice(args, "sound", PSoundChannel::Recorder))
    return FALSE;
  if (!SetSoundDevice(args, "sound", PSoundChannel::Player))
    return FALSE;
  if (!SetSoundDevice(args, "sound-in", PSoundChannel::Recorder))
    return FALSE;
  if (!SetSoundDevice(args, "sound-out", PSoundChannel::Player))
    return FALSE;

  noFastStart = args.HasOption('f');

  if (args.HasOption('u'))
    SetLocalUserName(args.GetOptionString('u'));

  if (args.HasOption('j')) {
    unsigned jitter = args.GetOptionString('j').AsUnsigned();
    if (jitter >= 20 && jitter <= 10000)
      SetMaxAudioDelayJitter(jitter);
    else {
      cout << "Jitter parameter should be between 20 milliseconds and 10 seconds" << endl;
      return FALSE;
    }
  }

  // Display configuration
  cout << "Local username: " << GetLocalUserName() << "\n"
          "Jitter buffer: "  << GetMaxAudioDelayJitter() << " ms\n"
          "Sound output device: \"" << GetSoundChannelPlayDevice() << "\"\n"
          "Sound  input device: \"" << GetSoundChannelRecordDevice() << "\"\n";

  InitialiseCodecs(args);

  cout <<  "Codecs (in preference order):\n";
  for (PINDEX i = 0; i < capabilityTable.GetSize(); i++)
    cout << "  " << capabilityTable[i] << '\n';
  cout << endl;

  if (args.HasOption('l')) {
    H323ListenerTCP * listener = new H323ListenerTCP(*this);
    if (!StartListener(listener)) {
      cout <<  "Could not open H.323 listener port on "
           << listener->GetListenerPort() << endl;
      return FALSE;
    }
  }

  if (args.HasOption('g')) {
    PString gkName = args.GetOptionString('g');
    if (SetGatekeeper("ip:"+gkName, new H323TransportUDP(*this)))
      cout << "Gatekeeper set: " << *gatekeeper << endl;
    else {
      cout << "Error registering with gatekeeper at \"" << gkName << '"' << endl;
      return FALSE;
    }
  }
  else if (!args.HasOption("no-gatekeeper")) {
    cout << "Searching for gatekeeper..." << flush;
    if (DiscoverGatekeeper(new H323TransportUDP(*this)))
      cout << "\nGatekeeper found: " << *gatekeeper << endl;
    else {
      cout << "\nNo gatekeeper found." << endl;
      if (args.HasOption("require-gatekeeper"))
        return FALSE;
    }
  }

  return TRUE;
}


BOOL MyH323EndPoint::InitialiseCodecs(PArgList & args)
{
  H323Capability * ulaw64 = new H323_G711Capability(H323_G711Capability::muLaw, H323_G711Capability::At64k);
  H323Capability * alaw64 = new H323_G711Capability(H323_G711Capability::ALaw, H323_G711Capability::At64k);
  H323Capability * gsm    = new H323_GSM0610Capability;
  //H323Capability * t120   = new H323_T120Capability;
  H323Capability * uic    = new H323_UserInputCapability;


  // The order in which capabilities are added to the capability table
  // determines which one is selected by default.

  if (args.HasOption("gsm"))
    AddCapability(gsm);
  if (args.HasOption("g711-ulaw"))
    AddCapability(ulaw64);
  if (args.HasOption("g711-Alaw"))
    AddCapability(alaw64);

  // Now add in the default order, if already added by above AddCapability
  // does nothing
  AddCapability(gsm);
  AddCapability(ulaw64);
  AddCapability(alaw64);
  //AddCodec(t120);
  AddCapability(uic);

  SetCapability(0, 0, gsm);
  SetCapability(0, 0, ulaw64);
  SetCapability(0, 0, alaw64);
  //SetCapability(0, 1, t120);
  SetCapability(0, 0, uic);

  return TRUE;
}


BOOL MyH323EndPoint::SetSoundDevice(PArgList & args,
                                    const char * optionName,
                                    PSoundChannel::Directions dir)
{
  if (!args.HasOption(optionName))
    return TRUE;

  PString dev = args.GetOptionString(optionName);

  if (dir == PSoundChannel::Player) {
    if (SetSoundChannelPlayDevice(dev))
      return TRUE;
  }
  else {
    if (SetSoundChannelRecordDevice(dev))
      return TRUE;
  }

  cout << "Device for " << optionName << " (\"" << dev << "\") must be one of:\n";

  PStringArray names = PSoundChannel::GetDeviceNames(dir);
  for (PINDEX i = 0; i < names.GetSize(); i++)
    cout << "  \"" << names[i] << "\"\n";

  return FALSE;
}


H323Connection * MyH323EndPoint::CreateConnection(unsigned callReference)
{
  return new MyH323Connection(*this, callReference, noFastStart);
}


void MyH323EndPoint::OnConnectionEstablished(H323Connection & connection,
                                             const PString & token)
{
  currentCallToken = token;
  cout << "Call with \"" << connection.GetRemotePartyName() << "\" established." << endl;
}


void MyH323EndPoint::OnConnectionCleared(H323Connection & connection, const PString &)
{
  cout << "Call with \"" << connection.GetRemotePartyName() << "\" terminated." << endl;
  if (uiState == uiMakingCall)
    exitFlag.Signal();
  currentCallToken = "";
}


PChannel * MyH323EndPoint::OpenAudioChannel(BOOL isEncoding, unsigned bufferSize)
{
  return H323EndPoint::OpenAudioChannel(isEncoding, bufferSize);
}


void MyH323EndPoint::ListenForIncomingCalls()
{
  cout << "Waiting for incoming calls for \"" << GetLocalUserName() << "\"\n";
  uiState = uiWaitingForCall;
}


void MyH323EndPoint::MakeOutgoingCall(const PString & dest, const PString & proxy)
{
  H323TransportTCP * transport = new H323TransportTCP(*this);

  if (!proxy) {
    if (!transport->SetRemoteAddress("ip:"+proxy)) {
      cout << "Invalid gateway address: \"" << proxy << '"' << endl;
      return;
    }
    cout << GetLocalUserName() << " is calling " << dest << '@' << proxy << endl;
  }
  else {
    if (!transport->SetRemoteAddress("ip:"+dest)) {
      cout << "Invalid destination address: \"" << dest << '"' << endl;
      return;
    }
    cout << GetLocalUserName() << " is calling host " << dest << endl;
  }

  if (!MakeCall(dest, transport, currentCallToken)) {
    cout << "Error making call to \"" << dest << '"' << endl;
    return;
  }

  uiState = uiMakingCall;
}


void MyH323EndPoint::AwaitTermination()
{
  new UserInterfaceThread(*this);
  exitFlag.Wait();

  if (!currentCallToken)
    ClearCall(currentCallToken);
}


void MyH323EndPoint::HandleUserInterface()
{
  PConsoleChannel console(PConsoleChannel::StandardInput);

  PTRACE(2, "Voxilla\tUser interface thread started.");

  for (;;) {
    cout << "Enter 0-9 for user indication, H to hang up, X to exit." << endl;

    if (isdigit(console.peek())) {
      H323Connection * connection = FindConnectionWithLock(currentCallToken);
      if (connection == NULL)
        cout << "No active connection\n";
      else {
        PString str;
        console >> str;
        cout << "Sending user indication: " << str << endl;
        connection->SendUserInput(str);
        connection->Unlock();
      }
    }
    else {
      switch (console.peek()) {
        case 'X' :
        case 'x' :
          cout << "Exiting." << endl;
          uiState = uiWaitingForCall;
          answerFlag.Signal();
          exitFlag.Signal();
          return;

        case 'H' :
        case 'h' :
          if (!currentCallToken) {
            cout << "Hanging up call." << endl;
            if (!ClearCall(currentCallToken))
              cout << "Could not hang up current call!\n";
          }
          break;

        case 'N' :
        case 'n' :
          if (uiState == uiAnsweringCall) {
            uiState = uiWaitingForCall;
            answerFlag.Signal();
          }
          break;

        default :
          if (uiState == uiAnsweringCall) {
            uiState = uiAnsweredCall;
            answerFlag.Signal();
          }
      }
    }

    console.ignore(INT_MAX, '\n');
  }
}


///////////////////////////////////////////////////////////////

MyH323Connection::MyH323Connection(MyH323EndPoint & ep,
                                   unsigned callReference,
                                   BOOL _noFastStart)
  : H323Connection(ep, callReference),
    myEndpoint(ep),
    noFastStart(_noFastStart)
{
}


BOOL MyH323Connection::OnAnswerCall(const PString & caller,
                                    const H323SignalPDU & /*setupPDU*/,
                                    H323SignalPDU & /*connectPDU*/)
{
  PTRACE(1, "H225\tOnWaitForAnswer - user input");

  cout << "Incoming call from \"" << caller << "\", Answer call (Y/n)? " << flush;

  myEndpoint.uiState = MyH323EndPoint::uiAnsweringCall;
  myEndpoint.answerFlag.Wait();

  if (myEndpoint.uiState != MyH323EndPoint::uiAnsweredCall) {
    cout << "Rejecting call." << endl;
    return FALSE;
  }

  myEndpoint.uiState = MyH323EndPoint::uiWaitingForCall;

  cout << "Accepting call." << endl;
  return TRUE;
}


void MyH323Connection::OnSelectLogicalChannels()
{
  if (fastStartState != FastStartDisabled && noFastStart)
    return;

#if 1 // Clear to temporarily disable the transmitter start up
  H323Connection::OnSelectLogicalChannels();
#endif
}


BOOL MyH323Connection::OnOpenLogicalChannel(const H245_OpenLogicalChannel & openPDU,
                                            H245_OpenLogicalChannelAck & ackPDU,
                                            unsigned & errorCode)
{
#if 0 // Set to temporarily disable the receiver acceptance
  return FALSE;
#else
  return H323Connection::OnOpenLogicalChannel(openPDU, ackPDU, errorCode);
#endif
}


BOOL MyH323Connection::OnStartLogicalChannel(H323Channel & channel)
{
  if (!H323Connection::OnStartLogicalChannel(channel))
    return FALSE;

  cout << "Started logical channel: ";

  switch (channel.GetDirection()) {
    case H323Channel::IsTransmitter :
      cout << "sending ";
      break;

    case H323Channel::IsReceiver :
      cout << "receiving ";
      break;

    default :
      break;
  }

  cout << channel.GetCapability() << endl;

  return TRUE;
}


BOOL MyH323Connection::OnAlerting(const H323SignalPDU & /*alertingPDU*/,
                                  const PString & username)
{
  cout << "Ringing phone for \"" << username << "\" ..." << endl;
  return TRUE;
}


void MyH323Connection::OnUserInputString(const PString & value)
{
  cout << "User input received: \"" << value << '"' << endl;
}


// End of File ///////////////////////////////////////////////////////////////
